#!/bin/env bash

BN=${0%%*/}

usage() {
  ( echo "Usage:"
    echo "  $BN -f file { [-s section] [-k key] } [-v value [-r subst] | -g] [-i]"
    echo "Options:"
    echo "  -f: INI file name"
    echo "  -s: section (REGEX) as in [section]"
    echo "  -k: key (REGEX) as in key="
    echo "  -v: set key to value"
    echo "  -r: replace subst (REGEX) with value in key"
    echo "  -g: get key value (first occurrence)"
    echo "  -i: edit file in place" ) 1>&2
  exit 1
}

[ $# -eq 0 ] && usage

unset FILE SECTION KEY VALUE SUBST GET INPLACE
while getopts f:s:k:v:r:gi c ; do
  case $c in
    f) FILE=$OPTARG ;;
    s) SECTION=$OPTARG ;;
    k) KEY=$OPTARG ;;
    v) VALUE=$OPTARG ;;
    r) SUBST=$OPTARG ;;
    g) GET=1 ;;
    i) INPLACE=1 ;;
    *) usage ;;
  esac
done

[ -z "$FILE" -o \( -z "$KEY" -a -z "$SECTION" \) ] && usage
[ -n "$VALUE" -a -n "$GET" ] && usage
[ -z "$VALUE" -a -n "$SUBST" ] && usage
[ ! -f "$FILE" ] && usage

EPOCH=`date "+%s"`
[ -z "$INPLACE" ] && FROM=$FILE || { FROM=$FILE.~$EPOCH~; cp -p $FILE $FROM; }

export SUBST SECTION KEY
awk -F= -v section="$SECTION" -v key="$KEY" -v set="$VALUE" -v get="$GET" -v subst="$SUBST" '
BEGIN{
  i = 0; spc = ""; spc_defd = 0;
  if (section ~ /^$/) i = 1;
} {
  # get leading space for keys if not yet defined
  if (!spc_defd && $1 !~ /^[[:space:]]*#/ && $2 ~ /./ && $2 ~ /[^[:space:]]/) { spc = $2; spc_defd = 1; sub(/[^[:space:]].*/, "", spc); }
  # if not section line and section found and ( key is undef or key matches )
  if ($0 !~ /^[[:space:]]*(\[|$)/ && i > 0 && (key ~ /^$/ || ( key ~ /./ && /^[[:space:]]*'"$KEY"'[[:space:]]*=/))) {
    # print line if get is defined but key not
    if (get ~ /./ && key ~ /^$/) { print $0; }
    # if get is defined (and key matches) print value
    else if (get ~ /./) { sub(/^[^=]*=[[:space:]]*/,""); print $0; exit 0; }
    # substitute SUBST in set
    else if (subst ~ /./) { gsub(/'"$SUBST"'/, set); print $0; }
    # print everything (used to add a new section)
    else if (key ~ /^$/) { print $0; }
    # set key to new value
    else { printf "%s=%s%s\n", $1, spc, set; }
    i = 2;
  } else {
    if (section ~ /./) {
      # set section found if section matches
      if (/^[[:space:]]*\['"$SECTION"'\]/) { i = 1; }
      # non-matching section
      else if (/^[[:space:]]*\[/) {
        # add new key to prior section
        if (get ~ /^$/ && key ~ /./ && i == 1 && subst ~ /^$/) printf "%s=%s%s\n", key, spc, set;
        # unset section found
        if (i > 0) i = -i;
      }
    }
    # print section line (if not yet done)
    if (get ~ /^$/) print $0;
  }
} END {
  # if section is empty and key not found
  if (section ~ /^$/ && i != 2) i = 0;
  if (get ~ /^$/ && (i == 0 || i == 1) && subst ~ /^$/) {
    # add new section if defined
    if (section ~ /./ && i == 0) printf "[%s]\n", section;   
    # add new key if defined
    if (key ~ /./) printf "%s=%s%s\n", key, spc, set;
  }
}' < $FROM | { [ -z "$INPLACE" ] && cat - || cat - > $FILE; }

[ -n "$INPLACE" ] && rm -f $FROM
exit 0
